/*
 * drivers/mtd/nand/ts7250.c
 *
 * Copyright (C) 2004 Technologic Systems (support@embeddedARM.com)
 *
 * Derived from drivers/mtd/nand/edb7312.c
 *   Copyright (C) 2004 Marius Gröger (mag@sysgo.de)
 *
 * Derived from drivers/mtd/nand/autcpu12.c
 *   Copyright (c) 2001 Thomas Gleixner (gleixner@autronix.de)
 *
 * $Id: tsprosoft.c,v 1.1 2007/11/15 23:18:09 ronald Exp $
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * Overview:
 *   This is a device driver for the NAND flash device found on the
 *   TS-7250 board which utilizes a Samsung 32 Mbyte part.
 */

#include <linux/slab.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/nand.h>
#include <linux/mtd/partitions.h>
#include <asm/io.h>
#include <asm/arch/hardware.h>
#include <asm/sizes.h>
#include <asm/mach-types.h>

struct partition_info { // little-endian format
  unsigned int bootable:8; // 0x80 = active, 0x00 = don't use for booting
  unsigned int starting_head:8;
  unsigned int starting_csect:8;
  unsigned int starting_xcyl:8;
  // MSB bits 0-5 are sector
  unsigned int system_id:8; // format of partition
  unsigned int ending_head:8;
  unsigned int ending_csect:8;
  unsigned int ending_xcyl:8;
  unsigned int relative_sector:32; // sector offset from start of disk to start of volume
  unsigned int total_sectors:32;
} __attribute__((packed));

struct partition_table {
  struct partition_info part[4];
} __attribute__((__packed__));

struct MBR {
  char bootcode[446];
  struct partition_table pt;
  unsigned int sig55:8;
  unsigned int sigAA:8;
} __attribute__((__packed__));

/*
 * MTD structure for TS7250 board
 */
static struct mtd_info *ts7250_mtd = NULL;

#ifdef CONFIG_MTD_PARTITIONS
static const char *part_probes[] = { "cmdlinepart", NULL };

#define NUM_PARTITIONS 3

/*
 * Define static partitions for flash device
 */
static struct mtd_partition partition_info32[] = {
	{
		.name		= "TS-BOOTROM",
		.offset		= 0x00000000,
		.size		= 0x00020000,
	}, {
		.name		= "Linux",
		.offset		= 0x00020000,
		.size		= 0x01d00000,
	}, {
		.name		= "RedBoot",
		.offset		= 0x01d20000,
		.size		= 0x002fc000,
	},
};

/*
 * Define static partitions for flash device
 */
static struct mtd_partition partition_info128[] = {
	{
		.name		= "TS-BOOTROM",
		.offset		= 0x00000000,
		.size		= 0x00020000,
	}, {
		.name		= "Linux",
		.offset		= 0x00020000,
		.size		= 0x07d00000,
	}, {
		.name		= "RedBoot",
		.offset		= 0x07d20000,
		.size		= 0x002e0000,
	},
};

static struct mtd_partition partition_info512[] = {
	{ .name= "whole chip",
		  .offset= 0,
		  .size= 0x20000000 },
	{ .name= "kernel",
		  .offset= 0x20000,
		  .size= 0x400000 },
	{ .name= "initrd",
		  .offset= 0x420000,
		  .size= 0x400000 },
	{ .name= "rootfs",
		  .offset= 0x820000,
		  .size= 0x20000000 - 0x820000 },
	{ .name= "part4",
		  .offset= 0x20000000,
		  .size= 0 },
};
#endif


/*
 *	hardware specific access to control-lines
 *
 *	ctrl:
 *	NAND_NCE: bit 0 -> bit 2
 *	NAND_CLE: bit 1 -> bit 1
 *	NAND_ALE: bit 2 -> bit 0
 */
static void ts7250_hwcontrol(struct mtd_info *mtd, int cmd, unsigned int ctrl)
{
	struct nand_chip *chip = mtd->priv;

	if (ctrl & NAND_CTRL_CHANGE) {
		unsigned long addr = TS72XX_NAND_CONTROL_VIRT_BASE;
		unsigned char bits;

		bits = (ctrl & NAND_NCE) << 2;
		bits |= ctrl & NAND_CLE;
		bits |= (ctrl & NAND_ALE) >> 2;

		__raw_writeb((__raw_readb(addr) & ~0x7) | bits, addr);
	}

	if (cmd != NAND_CMD_NONE)
		writeb(cmd, chip->IO_ADDR_W);
}

/*
 *	read device ready pin
 */
static int ts7250_device_ready(struct mtd_info *mtd)
{
	return __raw_readb(TS72XX_NAND_BUSY_VIRT_BASE) & 0x20;
}

/*
 * Main initialization routine
 */
static int __init ts7250_init(void)
{
	struct nand_chip *this;
	const char *part_type = 0;
	int mtd_parts_nb = 0;
        char buf[2048];
        int i,got = 0,parts=0;
        unsigned ofs=0;
        struct MBR *mbr = (struct MBR *)buf;

	if (!machine_is_ts72xx() || board_is_ts7200())
		return -ENXIO;

	/* Allocate memory for MTD device structure and private data */
	ts7250_mtd = kmalloc(sizeof(struct mtd_info) + sizeof(struct nand_chip), GFP_KERNEL);
	if (!ts7250_mtd) {
		printk("Unable to allocate TS7250 NAND MTD device structure.\n");
		return -ENOMEM;
	}

	/* Get pointer to private data */
	this = (struct nand_chip *)(&ts7250_mtd[1]);

	/* Initialize structures */
	memset(ts7250_mtd, 0, sizeof(struct mtd_info));
	memset(this, 0, sizeof(struct nand_chip));

	/* Link the private data with the MTD structure */
	ts7250_mtd->priv = this;
	ts7250_mtd->owner = THIS_MODULE;

	/* insert callbacks */
	this->IO_ADDR_R = (void *)TS72XX_NAND_DATA_VIRT_BASE;
	this->IO_ADDR_W = (void *)TS72XX_NAND_DATA_VIRT_BASE;
	this->cmd_ctrl = ts7250_hwcontrol;
	this->dev_ready = ts7250_device_ready;
	this->chip_delay = 15;
	this->ecc.mode = NAND_ECC_SOFT;

	printk("Searching for NAND flash...\n");
	/* Scan to find existence of the device */
	if (nand_scan(ts7250_mtd, 1)) {
		kfree(ts7250_mtd);
		return -ENXIO;
	}

	for (i=0;i<512;i++) {
	  buf[i] = 0;
	}
	ts7250_mtd->read(ts7250_mtd,0,512,&got,buf);

	if ( mbr->sigAA == 0xAA && mbr->sig55 == 0x55) {
          part_type = "MBR";
	  for (i=0;i<4;i++) {
	    if (mbr->pt.part[i].total_sectors) {
	      parts++;
	      partition_info512[i+1].offset = 
	        512 * mbr->pt.part[i].relative_sector;
	      if (i < 2) {
	        partition_info512[i+1].size = 
	          512 * mbr->pt.part[i+1].relative_sector -
	          512 * mbr->pt.part[i].relative_sector;
	      } else {
	        partition_info512[i+1].size = 
	          512 * mbr->pt.part[i].total_sectors;
	      }
	      ofs = partition_info512[i+1].offset 
	        + partition_info512[i+1].size;
	    }
	  }
	} else {
          part_type = "default";
	  parts = 3;
	}
	
	/* Register the partitions */
	printk(KERN_NOTICE "Using %s partition definition\n", part_type);
	add_mtd_partitions(ts7250_mtd, partition_info512, parts + 1);

	/* Return happy */
	return 0;
}

module_init(ts7250_init);

/*
 * Clean up routine
 */
static void __exit ts7250_cleanup(void)
{
	/* Unregister the device */
	del_mtd_device(ts7250_mtd);

	/* Free the MTD device structure */
	kfree(ts7250_mtd);
}

module_exit(ts7250_cleanup);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Jesse Off <joff@embeddedARM.com>");
MODULE_DESCRIPTION("MTD map driver for Technologic Systems TS-7250 board");
